//! 页面置换算法

use super::*;
use crate::memory::{frame::FrameTracker, *};
use alloc::collections::VecDeque;
use alloc::vec::Vec;
use core::borrow::Borrow;

/// 管理一个线程所映射的页面的置换操作
pub trait Swapper {
    /// 新建带有一个分配数量上限的置换器
    fn new(quota: usize) -> Self;

    /// 是否已达到上限
    fn full(&self) -> bool;

    /// 取出一组映射
    fn pop(&mut self) -> Option<(VirtualPageNumber, FrameTracker)>;

    /// 添加一组映射（不会在以达到分配上限时调用）
    fn push(&mut self, vpn: VirtualPageNumber, frame: FrameTracker, entry: *mut PageTableEntry);

    /// 只保留符合某种条件的条目（用于移除一段虚拟地址）
    fn retain(&mut self, predicate: impl Fn(&VirtualPageNumber) -> bool);
}

// pub type SwapperImpl = FIFOSwapper;

pub type SwapperImpl = ClockSwapper;

/// 页面置换算法基础实现：FIFO
pub struct FIFOSwapper {
    /// 记录映射和添加的顺序
    queue: VecDeque<(VirtualPageNumber, FrameTracker)>,
    /// 映射数量上限
    quota: usize,
}

impl Swapper for FIFOSwapper {
    fn new(quota: usize) -> Self {
        Self {
            queue: VecDeque::new(),
            quota,
        }
    }
    fn full(&self) -> bool {
        self.queue.len() == self.quota
    }
    fn pop(&mut self) -> Option<(VirtualPageNumber, FrameTracker)> {
        self.queue.pop_front()
    }
    fn push(&mut self, vpn: VirtualPageNumber, frame: FrameTracker, _entry: *mut PageTableEntry) {
        self.queue.push_back((vpn, frame));
    }
    fn retain(&mut self, predicate: impl Fn(&VirtualPageNumber) -> bool) {
        self.queue.retain(|(vpn, _)| predicate(vpn));
    }
}

// (a=0, m=0) 没有使用也没有修改，被逐出的优先级最高；
// (a=1, m=0) 使用过，但是没有修改过，优先级第二；
// (a=0, m=1) 没有使用过，但是修改过，优先级第三；
// (a=1, m=1) 使用过也修改过，优先级第四。
pub struct ClockSwapper {
    clock: Vec<ClockItem>,
    /// 映射数量上限
    quota: usize,
}
pub struct ClockItem {
    vpn: VirtualPageNumber,
    frame: FrameTracker,
    //优先级小的先淘汰
    priority: u8,
}

impl ClockItem {
    pub fn new(vpn: VirtualPageNumber, frame: FrameTracker, entry: *mut PageTableEntry) -> Self{
        let f = unsafe { entry.read() }.flags();
        let dirty = f.contains(Flags::DIRTY);
        let accessed = f.contains(Flags::ACCESSED);
        let mut priority = 0;
        if dirty  {
            priority += 2;
        }
        if accessed {
            priority += 1;
        }
        ClockItem{
            vpn,
            frame,
            priority
        }
    }
}
impl Swapper for ClockSwapper {
    fn new(quota: usize) -> Self {
        Self {
            clock: Vec::new(),
            quota,
        }
    }

    fn full(&self) -> bool {
        self.clock.len() == self.quota
    }

    //根据算法pop一个出来
    fn pop(&mut self) -> Option<(VirtualPageNumber, FrameTracker)> {
        let clock = &mut self.clock;
        let mut candidate: Option<(usize, u8)> = None;
        for i in 0..clock.capacity()-1 {
            if clock[i].priority == 0 {
                let ClockItem{ frame,vpn,.. } = clock.remove(i);
                return Some((vpn, frame));
            }
            let item = &clock[i];
            //没访问没修改，直接淘汰
            if let Some(cand) = candidate {
                //优先级更小
                if cand.1 > item.priority {
                    candidate = Some((i,item.priority));
                }
            } else if candidate.is_none() {
                candidate = Some((i,item.priority))
            }
        }
        if candidate.is_none() {
            None
        } else {
            let (idx, _) = candidate.unwrap();
            let ClockItem{ frame,vpn,.. } = clock.remove(idx);
            Some((vpn, frame))
        }
    }

    //不用判断是否已满
    fn push(&mut self, vpn: VirtualPageNumber, frame: FrameTracker, entry: *mut PageTableEntry) {
        let item = ClockItem::new(vpn,frame,entry);
        self.clock.push(item);
    }

    // 只保留符合某种条件的条目（用于移除一段虚拟地址）
    fn retain(&mut self, predicate: impl Fn(&VirtualPageNumber) -> bool) {
        self.clock.retain(|c| predicate(&c.vpn));
    }
}
