use crate::interrupt::{Context, timer};
use crate::sbi::console_getchar;
use riscv::register::scause::Scause;
use crate::PROCESSOR;
use crate::fs::STDIN;
use alloc::sync::Arc;
use crate::process::Thread;

/// 处理 ebreak 断点
///
/// 继续执行，其中 `sepc` 增加 2 字节，以跳过当前这条 `ebreak` 指令
pub fn breakpoint(context: &mut Context) -> *mut Context {
    println!("Breakpoint at 0x{:x}", context.sepc);
    context.sepc += 2;
    context
}

/// 处理时钟中断
pub fn supervisor_timer(context: &mut Context) -> *mut Context {
    timer::tick();
    PROCESSOR.lock().park_current_thread(context);
    PROCESSOR.lock().prepare_next_thread()
}

/// 处理外部中断，只实现了键盘输入
pub fn supervisor_external(context: &mut Context) -> *mut Context {
    let mut c = console_getchar();
    //处理sigint信号，它的值是3
    match c {
        3 => {
            //FIXME 对于user_shell程序，好像有点问题
            println!("terminated: killed by user");
            PROCESSOR.lock().kill_current_thread();
            return PROCESSOR.lock().prepare_next_thread();
        },
        99 => {
            println!("cloning thread");
            let current_thread: Arc<Thread> = PROCESSOR.lock().current_thread();
            //TODO clone 之后，程序出问题了
            let fork_thread = current_thread.fork(Some(*context)).expect("clone failed");
            PROCESSOR.lock().add_thread(fork_thread);
            println!("cloned current thread");
        },
        x if x < 255 => {
            if c == '\r' as usize {
                c = '\n' as usize;
            }
            STDIN.push(c as u8);
        }
        _ => {}
    }
    context
}

pub fn supervisor_soft(context: &mut Context) -> *mut Context { context }

/// 出现未能解决的异常，终止当前线程
pub fn fault(msg: &str, scause: Scause, stval: usize) -> *mut Context {
    println!(
        "{:#x?} terminated: {}",
        PROCESSOR.lock().current_thread(),
        msg
    );
    println!("cause: {:?}, stval: {:x}", scause.cause(), stval);

    PROCESSOR.lock().kill_current_thread();
    // 跳转到 PROCESSOR 调度的下一个线程
    PROCESSOR.lock().prepare_next_thread()
}