use crate::Scheduler;
use alloc::collections::LinkedList;
use alloc::collections::BinaryHeap;
use core::cmp::Ordering;
use core::cell::RefCell;
use core::num::Wrapping;

pub const BIG_STRIDE: usize = usize::MAX;
/// 将线程和调度信息打包
#[derive(Eq, PartialEq,Debug)]
pub struct StrideThread<ThreadType: Clone + Eq> {
    pass: RefCell<usize>,
    stride: RefCell<usize>,
    priority: RefCell<usize>,// >1
    /// 线程数据
    pub thread: ThreadType,
}

/// 采用Stride算法的调度器
#[derive(Debug)]
pub struct StrideScheduler<ThreadType: Clone + Eq> {
    /// 带有调度信息的线程池
    pool: BinaryHeap<StrideThread<ThreadType>>,
}
impl <ThreadType: Clone + Eq> StrideThread<ThreadType> {
    pub fn new(thread: ThreadType) -> Self {
        StrideThread{
            pass: RefCell::new(0),
            stride: RefCell::new(0),
            priority: RefCell::new(0),
            thread
        }
    }

}

impl <ThreadType: Clone + Eq> Ord for StrideThread<ThreadType> {
    fn cmp(&self, other: &Self) -> Ordering {
        //特殊的判断方法
        let res = (Wrapping(*self.pass.borrow()) - Wrapping(*other.pass.borrow())).0 as isize;
        //res > 0, 说明self比other大，我们要小的，所以返回Greater
        if res < 0 {
            Ordering::Greater
        }else if res == 0 {
            Ordering::Equal
        }else {
            Ordering::Less
        }
    }
}

impl <ThreadType: Clone + Eq> PartialOrd for StrideThread<ThreadType> {
    fn partial_cmp(&self, other: &Self) -> Option<Ordering> {
        Some(self.cmp(other))
    }
}

/// `Default` 创建一个空的调度器
impl <ThreadType: Clone + Eq> Default for StrideScheduler<ThreadType> {
    fn default() -> Self {
        Self {
            pool: BinaryHeap::new(),
        }
    }
}


pub type StridePriority = usize;

impl <ThreadType: Clone + Eq> Scheduler<ThreadType> for StrideScheduler<ThreadType> {
    type Priority = StridePriority;

    fn add_thread(&mut self, thread: ThreadType) {
        self.pool.push(StrideThread::new(thread));
    }

    fn get_next(&mut self) -> Option<ThreadType> {
        let item = self.pool.pop();
        if let Some(t) = item {
            {
                let mut pass = t.pass.borrow_mut();
                *pass = (Wrapping(*pass) + Wrapping(*t.stride.borrow())).0;
            }
            let thread = t.thread.clone();
            self.pool.push(t);
            Some(thread)
        } else {
            None
        }
    }

    fn remove_thread(&mut self, thread: &ThreadType) {
        //移除该线程
        self.pool.retain(|t| t.thread != *thread);
    }

    fn set_priority(&mut self, thread: ThreadType, priority: Self::Priority) {
        if priority < 2 {
            panic!("priority 必须 > 1");
        }
        let item = self.pool.iter().find(|t| t.thread == thread);
        let item = item.unwrap();
        let mut pri = item.priority.borrow_mut();
        //修改 优先级 和 步长
        *pri = priority;
        let mut stride = item.stride.borrow_mut();
        *stride = BIG_STRIDE / priority;
    }
}